# This code does spatial and non-spatial species distribution models for 
# 671 vascular plant species using four modelling methods and ensemble models.
# Code produces response curves and calculates variable importance for each method
# as well as evaluates the models.

# Analyses used in paper: Snow information is required in subcontinental scale predictions of mountain plant distributions
# By: Tuuli Rissanen, Pekka Niittynen, Janne Soininen and Miska Luoto
# Corresponding author: Tuuli Rissanen
# email: tuuli.rissanen@helsinki.fi
# Modified and deposited: 2021-04-14
#
########################################################

#Set the working directory
setwd(dirname(rstudioapi::getActiveDocumentContext()$path))
wd <- setwd(dirname(rstudioapi::getActiveDocumentContext()$path))
#getwd()

.libPaths(c(wd, .libPaths()))
libpath <- .libPaths()[1]

# Install newest version of sdm from github
# library(devtools, quiet = TRUE)
# install_github('babaknaimi/sdm',
#                upgrade = FALSE,
#                quiet = TRUE, 
#                lib = libpath)
library(sdm, quiet = TRUE)
library(raster, quiet = TRUE)
library(foreach, quiet = TRUE)
library(dplyr, quiet = TRUE)
library(dismo, quiet = TRUE)
library(randomForest, quiet = TRUE)
library(mgcv, quiet = TRUE)
library(ape, quiet = TRUE)
library(geosphere, quiet = TRUE)
library(spdep, quiet = TRUE)
library(gstat, quiet = TRUE)
library(tidyverse, quiet = TRUE)
library(data.table, quietly = TRUE)
library(doMPI, quietly = TRUE)

#####################
#ADD THE NEEDED DATA#
#####################

#Read the data
d <- as.data.frame(fread("./Data/AA_FennoSpecies.csv", header = T, dec = ".", sep = ",", encoding = "UTF-8"))
print(head(d))

d$bedrock <- as.factor(d$bedrock)
d$pf_rc <- as.factor(d$pf_rc)
d$name <- gsub(" ", "_", d$name) #change space in names to _
d$name <- gsub("-", "_", d$name) #change - to _

# Turn data to spatial points data frame (UTM34)
latlon <- "+proj=utm +zone=34 +datum=WGS84 +units=m +no_defs +ellps=WGS84 +towgs84=0,0,0" 
d2 <- d
coordinates(d2) <- ~x+y
crs(d2) <- latlon
rm(d)

# Read raster stack, covers Arctic-Alpine region of Fennoscandia
pred <- stack("./Data/AAModelStack_UTM34.tif")
names(pred) <- c("GDD3", "FDD", "WAB", "eledif", "bedrock", "snow", "pf_rc")
print(pred)

# Convert to Data frame
rdf <- as.data.frame(pred,xy=T)
# Set the same NA cells for every raster
pred[!complete.cases(rdf)] <- NA

#names(pred) <- read.csv("RasterNames.csv")$x[c(1,2,4,7,8)]
# Redo the converting with correct NAs
rdf <- as.data.frame(pred,xy=T)
# Remove NA rows
rdf <- rdf[complete.cases(rdf),]

#head(rdf)
#str(rdf)
rdf$bedrock <- as.factor(rdf$bedrock)
rdf$pf_rc <- as.factor(rdf$pf_rc)
#summary(rdf)
# Create an example raster as a template to the predictions
rtemp <- pred[[1]]
rtemp[!is.na(rtemp)] <- 1
# Remove the stack
rm(pred)

#Create a list of species names
#sort(table(d2$name))
species <- sort(unique(d2$name))

#Directory names for the result tables
resdir_eval <- "results_eval"
resdir_vi <- "results_varimp"
resdir_ens <- "results_ensemble"
resdir_MD <- "results_mindist"
resdir_curves <- "resp_curves"

#Directory names for the result maps (probability and binary)
mapdir_prob <- "maps_prob"
mapdir_bin <- "maps_bin"

# How many reps for each species and each modelling method
subreps <- 4

# Modelling methods
modmeths <- c('glm','gam','gbm','rf')

# The percentage of points to be used as a test data
testperce <- 30

# Let's model
if(!dir.exists(resdir_eval)){ dir.create(resdir_eval) }
if(!dir.exists(resdir_ens)){ dir.create(resdir_ens) }
if(!dir.exists(resdir_vi)){ dir.create(resdir_vi) }
if(!dir.exists(resdir_MD)){ dir.create(resdir_MD) }
if(!dir.exists(resdir_curves)){ dir.create(resdir_curves) }
if(!dir.exists(mapdir_bin)){ dir.create(mapdir_bin) }
if(!dir.exists(mapdir_prob)){ dir.create(mapdir_prob) }

#SDM-LOOP STARTS (RAC & NON-SPATIAL MODELS)
cl<-startMPIcluster()
registerDoMPI(cl)

foreach(i = species, .packages=c("raster","sdm","ape","mgcv","randomForest",
                                      "dismo","geosphere","spdep","gstat","dplyr")) %dopar% {
                                        #i <- species[1]
                                        #i <- "Empetrum_nigrum" #test species for RAC
                                        #i <- "Acer_platanoides"
                                        #Open pdf file for the response curves for each species
                                        pdf(paste0(resdir_curves,"/","Responses_",i,".pdf"), 10, 7)
                                        print(paste(i, Sys.time(), "Obs = ", NROW(d2[d2$name == i,])))
                                        
                                        tempd <- d2[d2$name == i,]
                                        # Background points from the cells that have at least one observation
                                        bgs <- d2[!d2$name == i,]
                                        bgs <- bgs[!duplicated(bgs$cellid),]
                                        bgs <- bgs[!bgs$cellid %in% tempd$cellid,]
                                        # At least 2000 bg points or the same number as presences
                                        bgs <- bgs[sample(1:NROW(bgs), ifelse(NROW(tempd) < 2000, 2000, NROW(tempd))),]
                                        bgs <- bgs[,-c(1)]
                                        
                                        
                                        # Create the modelling data
                                        sdmD1 <- sdmData(name~ GDD3 + FDD + WAB + eledif + f(bedrock) + snow + f(pf_rc),
                                                         train=tempd, bg = bgs)
                                        
                                        #Formula for the models
                                        
                                        f1 <- as.formula(paste0(i,"~GDD3 + FDD + WAB + eledif + f(bedrock) + snow + f(pf_rc)"))
                                        
                                        # Base models (train 100 % of the data, this is to get residuals for all points)
                                        # ennusteet myos tasta 
                                        m1 <- sdm(f1, data = sdmD1, methods=modmeths, replicatin='sub',
                                                  test.percent=0,n=1, 
                                                  setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                  modelSettings=list(gam=list(k=4, bs="tp"),
                                                                     gbm=list(n.trees = 500,train.fraction=1,
                                                                              interaction.depth=5,shrinkage=0.01,
                                                                              n.minobsinnode=5)),
                                                  parallelSettings=NULL)
                                        i1 <- getModelInfo(m1)
                                        i1 %>% filter(success) -> i1
                                        
                                        #Base model prediction to original data to calculate residuals
                                        p1 <- predict(m1,newdata=sdmD1@features)
                                        p1.1 <- as.data.frame(p1)
                                        colnames(p1.1) <- c("1","2","3","4")
                                        
                                        #######################################################
                                        #SAC & RAC TEST PART
                                        #########################################################
                                        
                                        #Calculating model residuals from sdm model prediction
                                        xy <- coordinates(m1@data) # xy coordinates
                                        id <- sort(m1@replicates[[i]][[1]]$train)
                                        moddata1 <- m1@models[[i]][[1]][[1]]@object$data 
                                        
                                        # Create list where to store the data
                                        dfs <- list()
                                        
                                        for(modelID in i1$modelID){
                                          
                                          #modelID <- 1
                                          
                                          tm <- m1@models[[i]][[i1$method[modelID]]][[i1$replicationID[modelID]]]
                                          gg <- data.frame(xy[id,], 
                                                           moddata1, 
                                                           predicted=p1.1[,modelID],
                                                           residuals=(moddata1[,i]- p1.1[,modelID])
                                          ) 
                                          # fitted = tm@object$fitted.values)
                                          
                                          dfs[[as.character(i1$method[modelID])]] <- gg
                                          
                                        }
                                        
                                        ###########################################################################
                                        # LOOP FOR THE INDIVIDUAL MODELLING METHODS
                                        # THE RAC MODEL PART STARTS
                                        ##########################################################################
                                        
                                        RAC_mods <- list() #list to store RAC models with cross validation (70 % 30 %)
                                        RAC_mods_full <- list() #list to store RAC models with no cross validation
                                        viAll_rac <- data.frame() #df for variable importance in RAC models
                                        evalAll_rac <- data.frame() #df for RAC model evaluations
                                        mindist_df <- data.frame()
                                        #THE RAC LOOP STARTS
                                        for(modelmethod in i1$method){
                                          #modelmethod <- "glm"
                                          # Extract the modelling data for wanted method
                                          moddata2 <- dfs[[modelmethod]]
                                          coordinates(moddata2) <- ~x+y
                                          crs(moddata2) <- latlon
                                          
                                          
                                          # Transforming the points to LatLon WGS84 because distance calculations need them in that format
                                          moddata2 <- spTransform(moddata2, "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs")
                                          
                                          # Calculate distance matrix between points
                                          dist.matrix <- distm(moddata2, fun=distGeo)
                                          
                                          # Testing when the SAC is dying away
                                          dist_bins <- 10000 # size of the distance classes in meters
                                          mindist <- 0
                                          repeat {
                                            print(paste0("Minimum distance = ", mindist))
                                            # Transform the distance matrix to binary with wanted distance thresholds
                                            dist.matrix.bin <- (dist.matrix > mindist & dist.matrix <= mindist + dist_bins)
                                            
                                            mi <- Moran.I(moddata2$residuals, dist.matrix.bin) # Calculate Morans' I with ape package
                                            
                                            print(paste0("observed Morans' I = ", mi$observed))
                                            print(paste0("p = ", mi$p.value))
                                            
                                            if (abs(mi$observed) < 0.05){ #Brakes when the Morans' I is less than the threshold
                                              break
                                            }
                                            mindist <- mindist+dist_bins
                                          }
                                          
                                          # The last distance with significant spatial autocorrelation
                                          if(mindist == 0 | mindist == dist_bins){
                                            mindist <- dist_bins
                                          } else {
                                            mindist <- mindist-dist_bins
                                          }
                                          
                                          # Save mindist in to a df
                                          MD <- as.data.frame(mindist)
                                          MD$model <- modelmethod
                                          MD$species <- i
                                          mindist_df <- rbind.data.frame(MD, mindist_df)
                                          
                                          # Calculating RAC
                                          gg_rac <- autocov_dist(z = moddata2$residuals, xy = moddata2, nbs=mindist/1000,
                                                                 style="B", type="inverse",zero.policy = TRUE)
                                       
                                          # Replace the residuals with RACs or add RACs as a new column
                                          moddata2$rac <- gg_rac
                                          
                                          #Formula for the models
                                          f2 <- as.formula(paste0(i,"~GDD3 + FDD + WAB + eledif + f(bedrock) + snow + f(pf_rc) + rac"))
                                          
                                          # New sdm-data
                                          sdmD2 <- sdmData(f2,train=moddata2)
                                          
                                          
                                          # Model with residual variable
                                          set.seed(123)
                                          m2 <- sdm(f2, data = sdmD2, methods=modelmethod, replicatin='sub',
                                                    test.percent=testperce, n=subreps, 
                                                    setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                    modelSettings=list(gam=list(k=4, bs="tp"),
                                                                       gbm=list(n.trees = 500,train.fraction=0.7,
                                                                                interaction.depth=5,shrinkage=0.01,
                                                                                n.minobsinnode=5)),
                                                    parallelSettings=NULL)
                                          i2 <- getModelInfo(m2)
                                          
                                          RAC_mods[[as.character(modelmethod)]] <- m2 #collect models into a list
                                          
                                          
                                          # Get the RAC model train and test evaluations
                                          etr2 <- getEvaluation(m2,w=1:NROW(i2),wtest='training',stat = c("AUC","COR","Deviance","TSS",
                                                                                                          "threshold","sensitivity",
                                                                                                          "specificity","Kappa"), opt = 2)
                                          etr2$mode <- "train"
                                          ete2 <- getEvaluation(m2,w=1:NROW(i2),wtest='test.dep',stat = c("AUC","COR","Deviance","TSS",
                                                                                                          "threshold","sensitivity",
                                                                                                          "specificity","Kappa"), opt = 2)
                                          ete2$mode <- "test"
                                          etr2[,1] <- ete2[,1] <- as.character(i2[,"method"])
                                          e2 <- rbind.data.frame(etr2, ete2)
                                          e2$species <- i
                                          evalAll_rac <- rbind.data.frame(evalAll_rac, e2)
                                          
                              
                                          # A MODEL WITH ALL DATA INCLUDED, i.e. NO CROSS VALIDATION
                                          # THIS IS USED FOR PREDICTIONS, VARIABLE IMPORTANCE & RESPONSE CURVES
                                          RAC_mods_full[[modelmethod]] <- sdm(f2, data = sdmD2, methods=modelmethod, replicatin='sub',
                                                                              test.percent=0, n=1, 
                                                                              setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                                              modelSettings=list(gam=list(k=4, bs="tp"),
                                                                                                 gbm=list(n.trees = 500,train.fraction=0.7,
                                                                                                          interaction.depth=5,shrinkage=0.01,
                                                                                                          n.minobsinnode=5)),
                                                                              parallelSettings=NULL)
                                          
                                          i2.1 <- getModelInfo(RAC_mods_full[[modelmethod]])
                                          
                                          #Variable importance from full model to have all the possible information
                                          for(ii in 1:NROW(i2.1)){
                                            #ii <- 1
                                            vi2 <- getVarImp(RAC_mods_full[[modelmethod]],id=ii,wtest='training')@varImportance
                                            vi2$species <- i
                                            vi2$model <- as.character(i2[ii,"method"])
                                            viAll_rac <- rbind.data.frame(viAll_rac, vi2)
                                            
                                          }
                                          
                                          
                                          #Response curves for each method from non cross validated model
                                          rcurve(RAC_mods_full[[modelmethod]], id=1:NROW(getModelInfo(RAC_mods_full[[modelmethod]])),
                                                 main = paste0("",i,"_",modelmethod, "_RAC")) 
                                          
                                        }
                                        
                                        # Write out minimum distances used in RAC calculations, 
                                        # RAC model evaluations and RAC model variable importance
                                        write.csv(mindist_df, paste0(resdir_MD,"/",i,"_mindist_SAC.csv"), row.names = F)
                                        write.csv(evalAll_rac, paste0(resdir_eval,"/",i,"_eval_RAC.csv"), row.names = F)
                                        write.csv(viAll_rac, paste0(resdir_vi,"/",i,"_varimp_RAC.csv"), row.names = F)
                                        
                                        #RAC loop ends
                                        #################
                                        
                                        #RAC ENSEMBLE MODELS AND PREDICTIONS
                                        
                                        # Loop over the modelling methods and make predictions to stack
                                        
                                        # Create dataframes where to combine the predictions
                                        # One for rac set as constant and other with interpolated rac
                                        tempdfmedian <- data.frame(reps = 1:NROW(rdf))
                                        tempdfidw <- data.frame(reps = 1:NROW(rdf))
                                        for(modelmethod in names(RAC_mods_full)){
                                          xyR <- as.data.frame(coordinates(RAC_mods_full[[modelmethod]]@data))
                                          modD <- RAC_mods_full[[modelmethod]]@data@features
                                          
                                          RACs <- as.data.frame(cbind(xyR,modD[,"rac"]))
                                          colnames(RACs) <- c("x","y","rac")
                                          
                                          coordinates(RACs) <- ~x+y
                                          crs(RACs) <- "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
                                          RACs <- spTransform(RACs, "+proj=utm +zone=34 +datum=WGS84 +units=m +no_defs +ellps=WGS84 +towgs84=0,0,0")
                                          
                                          # Inverse distance weighted interpolation with raster template
                                          gs <- gstat(formula = rac ~ 1, locations = RACs, nmax = 4, set = list(idp = 0.5))
                                          nn <- interpolate(rtemp, gs)
                                          
                                          # Mask
                                          nn <- mask(nn, rtemp)
                                          
                                          # Create dataframe for predictions
                                          rdfnn <- rdf
                                          rdfnn$rac <- na.omit(as.data.frame(nn, xy = F))[,1]
                                          
                                          
                                          head(rdf)
                                          rdf2 <- rdf
                                          rdf2$rac <- median(RACs$rac) #set original RAC values to median value
                                          
                                          #Predictions with median RAC value
                                          ensrac <- predict(RAC_mods_full[[modelmethod]], rdf2)
                                          
                                          # store to df
                                          tempdfmedian[,modelmethod] <- ensrac
                                          
                                          ######################################
                                          # predictions with interpolated RAC
                                          
                                          #Predictions
                                          ensrac <- predict(RAC_mods_full[[modelmethod]], rdfnn)
                                          tempdfidw[,modelmethod] <- ensrac
                                          
                                        }
                                        
                                        # Predictions to the original data
                                        tempdf <- data.frame(reps = 1:NROW(RAC_mods_full[[1]]@data@features))
                                        for(modelmethod in names(RAC_mods_full)){
                                          
                                          tempdf[,modelmethod] <- predict(RAC_mods_full[[modelmethod]],
                                                                          newdata=RAC_mods_full[[modelmethod]]@data@features)
                                          
                                        }
                                        
                                        
                                        # Create Ensemble model object from cv RAC models
                                        
                                        ensR <- RAC_mods[[1]]
                                        for(modelmethod in names(RAC_mods)[-1]){
                                          ensR <- ensR + RAC_mods[[modelmethod]]
                                        }
                                        # get evaluations from all the models
                                        eva <- getEvaluation(ensR,wtest='test.dep',stat = c("AUC","COR","Deviance","TSS",
                                                                                            "threshold","sensitivity",
                                                                                            "specificity","Kappa"), opt = 2)
                                        eva$model <- ensR@run.info$method
                                        # mean over modelling methods
                                        eva %>% group_by(model) %>% 
                                          summarise(TSS = mean(TSS)) -> evaw
                                        
                                        # Create ensemble prediction from full RAC mods. Model.meths. weighted by cross-validated TSS
                                        pt <- apply(tempdf[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                        # extract optimized Threshold to binarize
                                        resps <- c(rep(1, times=NROW(tempd)),rep(0, times=NROW(bgs)))
                                        trhsR <- evaluates(resps[RAC_mods_full[[1]]@replicates[[1]][[1]]$train],pt)@threshold_based[2,2]
                                        
                                        # Extraxt the predictions into the template raster
                                        # rac set as constant (median)
                                        ensRr <- rtemp
                                        ensRr[!is.na(ensRr)] <- apply(tempdfmedian[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                        ensRr <- ensRr*1000
                                        ensRr_prob <- ensRr
                                        #plot(ensRr_prob, main=paste0(i,"_RACmedian_prob"))
                                        writeRaster(ensRr_prob, paste0(mapdir_prob,"/",i,"_RACmedian_probability.tif"),
                                                    format = "GTiff", datatype = "INT2U", overwrite=T)
                                        
                                        # binary raster
                                        ensRr[ensRr < trhsR*1000] <- 0
                                        ensRr[ensRr > 0] <- 1
                                        ensRr_bin <- ensRr
                                        #plot(ensRr_bin, main=paste0(i,"_RACmedian_bin"))
                                        writeRaster(ensRr_bin, paste0(mapdir_bin,"/",i,"_RACmedian_binary.tif"),
                                                    format = "GTiff", datatype = "INT1U", overwrite=T)
                                        
                                        # Extraxt the predictions into the template raster
                                        # interpolated rac
                                        ensRr <- rtemp
                                        ensRr[!is.na(ensRr)] <- apply(tempdfidw[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                        ensRr <- ensRr*1000
                                        ensRr_prob <- ensRr
                                        #plot(ensRr_prob, main=paste0(i,"_RACidw_prob"))
                                        writeRaster(ensRr_prob, paste0(mapdir_prob,"/",i,"_RACidw_probability.tif"),
                                                    format = "GTiff", datatype = "INT2U", overwrite=T)
                                        # binary
                                        ensRr[ensRr < trhsR*1000] <- 0
                                        ensRr[ensRr > 0] <- 1
                                        ensRr_bin <- ensRr
                                        #plot(ensRr_bin, main=paste0(i,"_RACidw_bin"))
                                        writeRaster(ensRr_bin, paste0(mapdir_bin,"/",i,"_RACidw_binary.tif"),
                                                    format = "GTiff", datatype = "INT1U", overwrite=T)
                                        
                                        
                                        # EVALUATION STATISTICS FOR ENSEMBLE RAC MODELS with observed RAC
                                        # Predictive performance for these models is overly optimistic
                                        # because the data with observed rac is not completely independent but once seen by the models already.
                                        evaensR <- data.frame()
                                        for(ensreps in 1:subreps){
                                          #ensreps <- 1
                                          
                                          tempdf2 <- data.frame(reps = 1:NROW(RAC_mods[[1]]@data@features[RAC_mods[[1]]@replicates[[1]][[ensreps]]$test,]))
                                          for(modelmethod in names(RAC_mods)){
                                            
                                            pt <- predict(RAC_mods[[modelmethod]], run = ensreps,
                                                          newdata=RAC_mods[[modelmethod]]@data@features[RAC_mods[[modelmethod]]@replicates[[1]][[ensreps]]$test,])
                                            
                                            tempdf2[,modelmethod] <- pt
                                            
                                          }
                                          
                                          pt <- apply(tempdf2[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                          
                                          resps <- c(rep(1, times=NROW(tempd)),rep(0, times=NROW(bgs)))
                                          
                                          evaenstR <- evaluates(resps[RAC_mods[[1]]@replicates[[1]][[ensreps]]$test],pt)
                                          evaenstR <- data.frame(AUC = evaenstR@statistics$AUC,
                                                                 COR = evaenstR@statistics$COR[1],
                                                                 Deviance = evaenstR@statistics$Deviance,
                                                                 TSS = evaenstR@threshold_based[2,"TSS"],
                                                                 threshold = evaenstR@threshold_based[2,"threshold"],
                                                                 sensitivity = evaenstR@threshold_based[2,"sensitivity"],
                                                                 specificity = evaenstR@threshold_based[2,"specificity"],
                                                                 Kappa = evaenstR@threshold_based[2,"Kappa"])
                                          evaensR <- rbind.data.frame(evaensR, evaenstR)
                                          
                                        }
                                        write.csv(evaensR, paste0(resdir_ens,"/",i,"_RACobs_ensemble_eval.csv"), row.names = F)
                                        
                                        
                                        # EVALUATION STATISTICS FOR ENSEMBLE RAC MODELS with RAC set as constant (median)
                                        # This is to compare how the constant RAC is influencing the predictive performance
                                        evaensR <- data.frame()
                                        for(ensreps in 1:subreps){
                                          #ensreps <- 1
                                          
                                          tempdf2 <- data.frame(reps = 1:NROW(RAC_mods[[1]]@data@features[RAC_mods[[1]]@replicates[[1]][[ensreps]]$test,]))
                                          for(modelmethod in names(RAC_mods)){
                                            
                                            newd <- RAC_mods[[modelmethod]]@data@features[RAC_mods[[modelmethod]]@replicates[[1]][[ensreps]]$test,]
                                            newd$rac <- median(RAC_mods[[modelmethod]]@data@features[RAC_mods[[modelmethod]]@replicates[[1]][[ensreps]]$train,]$rac)
                                            pt <- predict(RAC_mods[[modelmethod]], run = ensreps,
                                                          newdata=newd)
                                            
                                            tempdf2[,modelmethod] <- pt
                                            
                                          }
                                          
                                          pt <- apply(tempdf2[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                          
                                          resps <- c(rep(1, times=NROW(tempd)),rep(0, times=NROW(bgs)))
                                          
                                          evaenstR <- evaluates(resps[RAC_mods[[1]]@replicates[[1]][[ensreps]]$test],pt)
                                          evaenstR <- data.frame(AUC = evaenstR@statistics$AUC,
                                                                 COR = evaenstR@statistics$COR[1],
                                                                 Deviance = evaenstR@statistics$Deviance,
                                                                 TSS = evaenstR@threshold_based[2,"TSS"],
                                                                 threshold = evaenstR@threshold_based[2,"threshold"],
                                                                 sensitivity = evaenstR@threshold_based[2,"sensitivity"],
                                                                 specificity = evaenstR@threshold_based[2,"specificity"],
                                                                 Kappa = evaenstR@threshold_based[2,"Kappa"])
                                          evaensR <- rbind.data.frame(evaensR, evaenstR)
                                          
                                        }
                                        write.csv(evaensR, paste0(resdir_ens,"/",i,"_RACmedian_ensemble_eval.csv"), row.names = F)
                                        
                                        
                                        
                                        ################################################################################
                                        # THE RAC MODEL PART ENDS
                                        ##################################################################################
                                        
                                        #################################################################################
                                        # THE NON-SPATIAL SDM PART BEGINS
                                        #################################################################################
                                        # the original data for species i is still the same as in the very begin
                                        # this is to ensure we do all modelling with the same train and test points
                                        # tempd and bgs are the same as earlier
                                        #sdm data = sdmD1
                                        # model formula = f1
                                        
                                        #i <- "Quercus_robur"
                                        print(paste(i, Sys.time(), "Obs = ", NROW(d2[d2$name == i,])))
                                        
                                        # Create the modelling data
                                        #sdmD1 <- sdmData(name~ GDD3 + FDD + WAB + eledif + f(bedrock) + snow + f(pf_rc),
                                        #                 train=tempd, bg = bgs)
                                        
                                        #Formula for the models
                                        # f1 <- as.formula(paste0(i,"~GDD3 + FDD + WAB + eledif + f(bedrock) + snow + f(pf_rc)"))
                                        
                                        #Full model with 100 % train to calculate variable importance and response curves
                                        # this is the same as in the very beginning
                                        #m1 <- sdm(f1, data = sdmD1, methods=modmeths, replicatin='sub',
                                        # test.percent=0,n=1, 
                                        #setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                        # modelSettings=list(gam=list(k=4, bs="ts"),
                                        #                    gbm=list(n.trees = 500,train.fraction=1,
                                        #                             interaction.depth=5,shrinkage=0.01,
                                        #                             n.minobsinnode=5)),
                                        #  parallelSettings=NULL)
                                        
                                        #Loop for variable importance for non-spatial sdms
                                        viAll <- data.frame()
                                        
                                        for(modelID in 1:NROW(i1)){
                                          
                                          #ii <- 1
                                          vi <- getVarImp(m1,id=modelID,wtest='training')@varImportance
                                          vi$species <- i
                                          vi$model <- as.character(i1[modelID,"method"])
                                          viAll <- rbind.data.frame(viAll, vi)
                                        }
                                        
                                        write.csv(viAll, paste0(resdir_vi,"/",i,"_varimp_nonS.csv"), row.names = F)
                                        
                                        # Response curves of non-spatial models
                                        
                                        #rcurve(m1, id=1:NROW(getModelInfo(m1)), main = paste0(i,"_nonS"))  
                                        rcurve(m1, id=1, main = paste0(i,"_glm_nonS"))  
                                        rcurve(m1, id=2, main = paste0(i,"_gam_nonS"))  
                                        rcurve(m1, id=3, main = paste0(i,"_gbm_nonS"))  
                                        rcurve(m1, id=4, main = paste0(i,"_rf_nonS"))  
                                        #rcurve(m1, method='gam',smooth=T, main = paste0(i,"_nonS"))  
                                        #rcurve(m1, method='gbm',smooth=T, main = paste0(i,"_nonS"))  
                                        
                                        
                                        # Models to calculate evaluation statistics (train 70 % and test 30 % of the data)
                                        set.seed(123)
                                        m3 <- sdm(f1, data = sdmD1, methods=modmeths, replicatin='sub',
                                                  test.percent=testperce,n=subreps, 
                                                  setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                  modelSettings=list(gam=list(k=4, bs='tp'),
                                                                     gbm=list(n.trees = 500,train.fraction=1,
                                                                              interaction.depth=5,shrinkage=0.01,
                                                                              n.minobsinnode=5)),
                                                  parallelSettings=NULL)
                                        i3 <- getModelInfo(m3)
                                        
                                        
                                        # Get train and test evaluations for non-spatial sdms
                                        etr <- getEvaluation(m3,w=1:NROW(i3),wtest='training',stat = c("AUC","COR","Deviance","TSS",
                                                                                                       "threshold","sensitivity",
                                                                                                       "specificity","Kappa"), opt = 2)
                                        etr$mode <- "train"
                                        ete <- getEvaluation(m3,w=1:NROW(i3),wtest='test.dep',stat = c("AUC","COR","Deviance","TSS",
                                                                                                       "threshold","sensitivity",
                                                                                                       "specificity","Kappa"), opt = 2)
                                        ete$mode <- "test"
                                        etr[,1] <- ete[,1] <- as.character(i3[,"method"])
                                        e <- rbind.data.frame(etr, ete)
                                        write.csv(e, paste0(resdir_eval,"/",i,"_eval_nonS.csv"), row.names = F)
                                        
                                        
                                        #NON-SPATIAL ENSEMBLE MODELS
                                        
                                        # Calculate test statistics for the ensemble models
                                        evaens <- data.frame()
                                        for(ensreps in 1:subreps){
                                          pt <- ensemble(m3,newdata=sdmD1@features[m3@replicates[[1]][[ensreps]]$test,],
                                                         setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep",
                                                                      id=seq(ensreps,length(modmeths)*subreps,by = subreps)))
                                          resps <- c(rep(1, times=NROW(tempd)),rep(0, times=NROW(bgs)))
                                          evaenst <- evaluates(resps[m3@replicates[[1]][[ensreps]]$test],pt)
                                          evaenst <- data.frame(AUC = evaenst@statistics$AUC,
                                                                COR = evaenst@statistics$COR[1],
                                                                Deviance = evaenst@statistics$Deviance,
                                                                TSS = evaenst@threshold_based[2,"TSS"],
                                                                threshold = evaenst@threshold_based[2,"threshold"],
                                                                sensitivity = evaenst@threshold_based[2,"sensitivity"],
                                                                specificity = evaenst@threshold_based[2,"specificity"],
                                                                Kappa = evaenst@threshold_based[2,"Kappa"])
                                          evaens <- rbind.data.frame(evaens, evaenst)
                                        }
                                        write.csv(evaens, paste0(resdir_ens,"/",i,"_nonS_ensemble_eval.csv"), row.names = F)
                                        
                                        # ENSEMBLE MAPS
                                        # Non-spatial models with cross validated TSS (need to be done like in RAC models
                                        NonS_mods_full <- list() #list to store NonS models with no cross validation
                                        
                                        for(modelmethod in modmeths){
                                          NonS_mods_full[[modelmethod]] <- sdm(f1, data = sdmD1, methods=modelmethod, replicatin='sub',
                                                                          test.percent=0,n=1, 
                                                                          setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                                          modelSettings=list(gam=list(k=4, bs="tp"),
                                                                                             gbm=list(n.trees = 500,train.fraction=1,
                                                                                                      interaction.depth=5,shrinkage=0.01,
                                                                                                      n.minobsinnode=5)),
                                                                          parallelSettings=NULL)
                                          
                                          i.1 <- getModelInfo(NonS_mods_full[[modelmethod]])
                                        }
                                        
                                        NonS_mods <- list() #list to store NonS models with cross validation (70 % 30 %)
                                        for(modelmethod in modmeths){
                                          NonS_mods[[modelmethod]] <- sdm(f1, data = sdmD1, methods=modelmethod, replicatin='sub',
                                                                     test.percent=testperce,n=subreps, 
                                                                     setting=list(method='weighted',stat='TSS',opt=2,wtest="test.dep"),
                                                                     modelSettings=list(gam=list(k=4, bs='tp'),
                                                                                        gbm=list(n.trees = 500,train.fraction=1,
                                                                                                 interaction.depth=5,shrinkage=0.01,
                                                                                                 n.minobsinnode=5)),
                                                                     parallelSettings=NULL)
                                        }
                                        
                                        
                                        tempdfmedian <- data.frame(reps = 1:NROW(rdf))
                                        #Predictions to whole area
                                        for(modelmethod in names(NonS_mods_full)){
                                          ens_all <- predict(NonS_mods_full[[modelmethod]], rdf)
                                          
                                          # store to df
                                          tempdfmedian[,modelmethod] <- ens_all
                                        }
                                        
                                        # Predictions to the original data
                                        tempdf <- data.frame(reps = 1:NROW(NonS_mods_full[[1]]@data@features))
                                        for(modelmethod in names(NonS_mods_full)){
                                          
                                          tempdf[,modelmethod] <- predict(NonS_mods_full[[modelmethod]],
                                                                          newdata=NonS_mods_full[[modelmethod]]@data@features)
                                          
                                        }
                                        
                                        # Create Ensemble model object from cross-validated NonS models
                                        ensR <- NonS_mods[[1]]
                                        ensR <- ensR + NonS_mods[[2]] + NonS_mods[[3]] + NonS_mods[[4]]
                                        
                                        for(modelmethod in names(NonS_mods)[-1]){
                                          ensR <- ensR + NonS_mods[[modelmethod]]
                                        } 
                                        
                                        i_ens <- getModelInfo(ensR)
                                        models_testdep <- as.data.frame(i_ens[i_ens$test.dep =="TRUE",c(1,3)])
                                        
                                        # get evaluations from all the models
                                        eva <- getEvaluation(ensR,wtest='test.dep',stat = c("AUC","COR","Deviance","TSS",
                                                                                            "threshold","sensitivity",
                                                                                            "specificity","Kappa"), opt = 2)
                                        
                                        eva$model <- ensR@run.info$method
                                        # mean over modelling methods
                                        eva %>% group_by(model) %>% 
                                          summarise(TSS = mean(TSS)) -> evaw
                                        
                                        # Create ensemble prediction from full RAC mods. Model.meths. weighted by cross-validated TSS
                                        pt <- apply(tempdf[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                        
                                        # extract optimized Threshold to binarize
                                        resps <- c(rep(1, times=NROW(tempd)),rep(0, times=NROW(bgs)))
                                        trhsR <- evaluates(resps[NonS_mods_full[[1]]@replicates[[1]][[1]]$train],pt)@threshold_based[2,2]
                                        
                                        # Extract the predictions into the template raster
                                        ensRr <- rtemp
                                        ensRr[!is.na(ensRr)] <- apply(tempdfmedian[,-1], 1, function(x) weighted.mean(x, w = evaw$TSS))
                                        ensRr <- ensRr*1000
                                        ensRr_prob <- ensRr
                                        #plot(ensRr_prob, col=viridis(50,option = "E",direction=-1), main=paste0(i,"_NonS_prob"))
                                        writeRaster(ensRr_prob, paste0(mapdir_prob,"/",i,"_NonS_probability.tif"),
                                                    format = "GTiff", datatype = "INT2U", overwrite=T)
                                        
                                        # binary
                                        ensRr[ensRr < trhsR*1000] <- 0
                                        ensRr[ensRr > 0] <- 1
                                        ensRr_bin <- ensRr
                                        #plot(ensRr_bin, main=paste0(i,"_NonS_bin"))
                                        writeRaster(ensRr_bin, paste0(mapdir_bin,"/",i,"_NonS_binary.tif"),
                                                    format = "GTiff", datatype = "INT1U", overwrite=T)
                                        
                                        
                                        # Remove the grd files from the working directory
                                        grds <- list.files(pattern = ".grd$")
                                        unlink(grds, recursive=TRUE, force=FALSE)
                                        grds2 <- list.files(path="Z:/Documents",pattern = ".grd$")
                                        unlink(grds2, recursive=TRUE, force=FALSE)
                                        
                                        # In the end this closes and writes the pdf file
                                        dev.off()
                                      }

closeCluster(cl)
mpi.quit()
